#include "mainwindow.h"
#include "post_to_main_thread.h"
#include "ui_mainwindow.h"

#include <stdlib.h>

using namespace std;

string escape_html(const string &str)
{
	string s = "";
	for (char ch : str) {
		if (ch == '<') {
			s += "&lt;";
		} else if (ch == '>') {
			s += "&gt;";
		} else if (ch == '&') {
			s += "&amp;";
		} else {
			s += ch;
		}
	}
	return s;
}

string escape_unicode(const string &str)
{
	string s = "";
	for (size_t pos = 0; pos < str.size(); ) {
		wchar_t wc;
		int len = mbtowc(&wc, str.data() + pos, str.size() - pos);
		if (len == -1) {
			wc = '?';
			len = 1;
		}
		pos += len;

		if (wc == '\\') {
			s += "\\\\";
		} else if (isprint(wc)) {
			s += wc;
		} else {
			char buf[16];
			snprintf(buf, sizeof(buf), "\\u%04x", wc);
			s += buf;
		}
	}
	return s;
}

string escape_quotes(const string &str)
{
	string s = "";
	for (char ch : str) {
		if (ch == '"') {
			s += '\\';
		}
		s += ch;
	}
	return s;
}

string serialize_as_json(const map<string, string> &param)
{
	string s = "{";

	bool first = true;
	for (const auto &key_value : param) {
		if (!first) s += ", ";
		first = false;

		s += '"';
		s += escape_quotes(escape_unicode(key_value.first));
		s += "\": \"";
		s += escape_quotes(escape_unicode(key_value.second));
		s += '"';
	}
	s += "}";
	return s;	
}

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
	ui->setupUi(this);
	ws = new WSServer("127.0.0.1", 5250);
	ws->set_connection_callback([this](bool connected) {
		string msg = connected ? "Connected" : "Not connected";
		post_to_main_thread([this, msg]() {
			ui->ws_connected_label->setText(QString::fromStdString(msg));
		});
	});

	connect(ui->ws_disconnect_btn, &QPushButton::clicked, this, &MainWindow::ws_disconnect_clicked);
	connect(ui->set_initials_btn, &QPushButton::clicked, this, &MainWindow::set_initials_clicked);
	connect(ui->set_color_btn, &QPushButton::clicked, this, &MainWindow::set_color_clicked);
	connect(ui->set_score_btn, &QPushButton::clicked, this, &MainWindow::set_score_clicked);
	connect(ui->set_all_scorebug_btn, &QPushButton::clicked, this, &MainWindow::set_all_scorebug_clicked);
	connect(ui->goal_1_btn, &QPushButton::clicked, this, [this]() { add_goal(ui->score_1_box, 1); });
	connect(ui->ungoal_1_btn, &QPushButton::clicked, this, [this]() { add_goal(ui->score_1_box, -1); });
	connect(ui->goal_2_btn, &QPushButton::clicked, this, [this]() { add_goal(ui->score_2_box, 1); });
	connect(ui->ungoal_2_btn, &QPushButton::clicked, this, [this]() { add_goal(ui->score_2_box, -1); });

	connect(ui->set_clock_btn, &QPushButton::clicked, this, &MainWindow::set_clock_clicked);
	connect(ui->set_clock_limit_btn, &QPushButton::clicked, this, &MainWindow::set_clock_limit_clicked);
	connect(ui->start_and_show_clock_btn, &QPushButton::clicked, this, &MainWindow::start_and_show_clock_clicked);
	connect(ui->stop_clock_btn, &QPushButton::clicked, this, &MainWindow::stop_clock_clicked);
	connect(ui->show_clock_btn, &QPushButton::clicked, this, &MainWindow::show_clock_clicked);
	connect(ui->hide_clock_btn, &QPushButton::clicked, this, &MainWindow::hide_clock_clicked);

	connect(ui->set_comment_btn, &QPushButton::clicked, this, &MainWindow::set_comment_clicked);
	connect(ui->set_and_show_comment_btn, &QPushButton::clicked, this, &MainWindow::set_and_show_comment_clicked);
	connect(ui->hide_comment_btn, &QPushButton::clicked, this, &MainWindow::hide_comment_clicked);
	connect(ui->set_and_show_autocomment_btn, &QPushButton::clicked, this, &MainWindow::set_and_show_autocomment_clicked);

	connect(ui->show_lower_third_btn, &QPushButton::clicked, this, &MainWindow::show_lower_third_clicked);
	connect(ui->hide_lower_third_btn, &QPushButton::clicked, this, &MainWindow::hide_lower_third_clicked);

	connect(ui->quick_lower_third_edit, &QLineEdit::returnPressed, this, &MainWindow::quick_lower_third_activate);
	connect(ui->show_quick_lower_third_btn, &QPushButton::clicked, this, &MainWindow::quick_lower_third_activate);

	connect(ui->show_scorebug_btn, &QPushButton::clicked, this, &MainWindow::show_scorebug_clicked);
	connect(ui->show_group_a_btn, &QPushButton::clicked, this, [this]() { show_group_clicked("Group A"); });
	connect(ui->show_group_b_btn, &QPushButton::clicked, this, [this]() { show_group_clicked("Group B"); });
	connect(ui->show_schedule_btn, &QPushButton::clicked, this, &MainWindow::show_schedule_clicked);
	connect(ui->show_carousel_btn, &QPushButton::clicked, this, &MainWindow::show_carousel_clicked);
	connect(ui->show_nothing_btn, &QPushButton::clicked, this, &MainWindow::show_nothing_clicked);
	connect(ui->show_roster_1_btn, &QPushButton::clicked, this, [this]() { show_roster_clicked(ui->initials_1_edit->text().toStdString()); });
	connect(ui->show_roster_2_btn, &QPushButton::clicked, this, [this]() { show_roster_clicked(ui->initials_2_edit->text().toStdString()); });
	connect(ui->show_roster_carousel_btn, &QPushButton::clicked, this, &MainWindow::show_roster_carousel_clicked);

	autocomment_update();

	const set<pair<unsigned, unsigned>> usb{{ 0x0e8f, 0x0041 }};
	event_device = new EventDevice(usb, ui->quick_lower_third_edit);
	event_device->start_thread();
}

MainWindow::~MainWindow()
{
	delete ui;
}

void MainWindow::ws_disconnect_clicked()
{
	ws->change_port(stoi(ui->ws_port_box->text().toStdString()));
}

void MainWindow::set_initials_clicked()
{
	map<string, string> param;
	param["team1"] = escape_html(ui->initials_1_edit->text().toStdString());
	param["team2"] = escape_html(ui->initials_2_edit->text().toStdString());
	ws->send_command("update " + serialize_as_json(param));
	ws->send_command("eval setteams()");
}

void MainWindow::set_color_clicked()
{
	map<string, string> param;
	param["team1color"] = ui->color_1_edit->text().toStdString();  // Should maybe be escaped, but meh.
	param["team2color"] = ui->color_2_edit->text().toStdString();
	ws->send_command("update " + serialize_as_json(param));
	ws->send_command("eval setcolors()");
}

void MainWindow::set_score_clicked()
{
	map<string, string> param;
	param["score1"] = to_string(ui->score_1_box->value());
	param["score2"] = to_string(ui->score_2_box->value());
	ws->send_command("update " + serialize_as_json(param));
	ws->send_command("eval setscore()");
	autocomment_update();
}

void MainWindow::set_all_scorebug_clicked()
{
	set_initials_clicked();
	set_color_clicked();
	set_score_clicked();
}

void MainWindow::add_goal(QSpinBox *box, int delta)
{
	box->setValue(box->value() + delta);
	set_score_clicked();
}

void MainWindow::set_clock_clicked()
{
	map<string, string> param;
	param["clock_min"] = to_string(ui->clock_min_box->value());
	param["clock_sec"] = to_string(ui->clock_sec_box->value());
	ws->send_command("update " + serialize_as_json(param));
	ws->send_command("eval setclockfromstate()");
}

void MainWindow::set_clock_limit_clicked()
{
	map<string, string> param;
	param["clock_limit_min"] = to_string(ui->clock_limit_min_box->value());
	param["clock_limit_sec"] = to_string(ui->clock_limit_sec_box->value());
	ws->send_command("update " + serialize_as_json(param));
	ws->send_command("eval setclocklimitfromstate()");
}

void MainWindow::start_and_show_clock_clicked()
{
	ws->send_command("eval startclock()");  // Also shows.
}

void MainWindow::stop_clock_clicked()
{
	ws->send_command("eval stopclock()");
}

void MainWindow::show_clock_clicked()
{
	ws->send_command("eval showclock()");
}

void MainWindow::hide_clock_clicked()
{
	ws->send_command("eval hideclock()");
}

void MainWindow::set_comment_clicked()
{
	map<string, string> param;
	param["comment"] = ui->comment_edit->text().toStdString();
	ws->send_command("update " + serialize_as_json(param));
	ws->send_command("eval setcomment()");
}

void MainWindow::set_and_show_comment_clicked()
{
	set_comment_clicked();
	ws->send_command("eval showcomment()");
}

void MainWindow::set_and_show_autocomment_clicked()
{
	ui->comment_edit->setText(ui->autocomment_edit->text());
	set_and_show_comment_clicked();
}

void MainWindow::hide_comment_clicked()
{
	ws->send_command("eval hidecomment()");
}

void MainWindow::show_lower_third_clicked()
{
	map<string, string> param;
	param["text1"] = ui->lowerthird_heading_edit->text().toStdString();
	param["text2"] = ui->lowerthird_subheading_edit->text().toStdString();
	ws->send_command("update " + serialize_as_json(param));
	ws->send_command("eval setandshowlowerthird()");
}

void MainWindow::hide_lower_third_clicked()
{
	ws->send_command("eval hidelowerthird()");
}

void MainWindow::quick_lower_third_activate()
{
	string code = ui->quick_lower_third_edit->text().toUpper().toStdString();
	if (code == "A") {
		add_goal(ui->score_1_box, 1);
	} else if (code == "B") {
		add_goal(ui->score_2_box, 1);
	} else if (code == "C") {
		ws->send_command("eval hidelowerthird()");
	} else {
		map<string, string> param;
		param["code"] = code;
		ws->send_command("update " + serialize_as_json(param));
		ws->send_command("eval quicklowerthird()");
	}
	ui->quick_lower_third_edit->clear();
}

void MainWindow::autocomment_update()
{
	int score1 = ui->score_1_box->value();
	int score2 = ui->score_2_box->value();
	string msg;
	if (abs(score1 - score2) >= 3) {
		msg = "Game ends after this point";
	} else {
		int cap = max(score1, score2) + 1;
		if (score1 == score2) ++cap;

		if (cap >= 13) {
			msg = "Point cap: First to 13";
		} else {
			char buf[32];
			snprintf(buf, sizeof(buf), "Pagacap: First to %d", cap);
			msg = buf;
		}
	}
	ui->autocomment_edit->setText(QString::fromStdString(msg));
}

void MainWindow::show_scorebug_clicked()
{
	ws->send_command("eval stopcarousel()");
	ws->send_command("eval hidetable()");
	ws->send_command("eval showscorebug()");
}

void MainWindow::show_group_clicked(const std::string &group_name)
{
	map<string, string> param;
	param["group_name"] = group_name;
	ws->send_command("eval stopcarousel()");
	ws->send_command("update " + serialize_as_json(param));
	ws->send_command("eval showgroup_from_state()");
}

void MainWindow::show_roster_clicked(const std::string &team_code)
{
	map<string, string> param;
	param["team_code"] = team_code;
	ws->send_command("eval stopcarousel()");
	ws->send_command("update " + serialize_as_json(param));
	ws->send_command("eval showroster_from_state()");
}

void MainWindow::show_schedule_clicked()
{
	ws->send_command("eval stopcarousel()");
	ws->send_command("eval showschedule()");
}

void MainWindow::show_carousel_clicked()
{
	ws->send_command("eval stopcarousel()");
	ws->send_command("eval showcarousel()");
}

void MainWindow::show_roster_carousel_clicked()
{
	map<string, string> param;
	param["team1"] = escape_html(ui->initials_1_edit->text().toStdString());
	param["team2"] = escape_html(ui->initials_2_edit->text().toStdString());
	ws->send_command("eval stopcarousel()");
	ws->send_command("update " + serialize_as_json(param));
	ws->send_command("eval showrostercarousel_from_state()");
}

void MainWindow::show_nothing_clicked()
{
	ws->send_command("eval hidescorebug()");
	ws->send_command("eval stopcarousel()");
	ws->send_command("eval hidetable()");
}
